"""
Public classes and functions of ml_tools subpackage
"""

from .truth_transformations import Mask2TwoClass, Mask2Label
from .classifier import ImageBaseClassifier, ImagePixelClassifier, ImagePatchClassifier, ImagePixel2PatchClassifier, \
    ImageClassificationMaskTask
from .validator import SGMLBaseValidator
from .postprocessing import MorphologicalOperations, MorphologicalStructFactory, PostprocessingTask,\
    MorphologicalFilterTask
from .train_test_split import TrainTestSplitTask

__version__ = '0.7.3'



"""
Module for transforming reference labels

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import numpy as np

# pylint: disable=invalid-name


class Mask2Label:
    """
    Transforms mask (shape n x m) into a single label. Transformation works in two modes:
        - majority: assign label to the class with the largest contribution
        - target: assign label to target if its contribution percentage is above or equal to target threshold.
        In other cases label is set to 0.

    The parameters target_value and target_threshold are taken into account only in target mode.

    Parameters
    ----------
    mode: str ('majority', 'target')
        conversion mode

    target_value: int (default 1)
        target value

    target_threshold: float (default 0.5)
        fraction of pixels in mask that need to belong to the target to be labeled as target
    """

    def __init__(self, mode, target_value=1, target_threshold=0.5):

        self.mode_ = mode
        if self.mode_ not in ['majority', 'target']:
            print('Invalid mode! Set mode to majority or target.')

        self.target_ = target_value
        self.threshold_ = target_threshold

    def _target(self, mask):
        unique, counts = np.unique(mask, return_counts=True)
        value_count = dict(zip(unique, counts))

        return 1 if self.target_ in value_count.keys() and \
                    value_count[self.target_] / np.ma.size(mask) >= self.threshold_ else 0

    @staticmethod
    def _majority(mask):
        label, count = np.unique(mask, return_counts=True)

        return label[np.argmax(count)]

    def transform(self, X):
        """
        Parameters
        ----------
        X : array-like, shape [n x m]
            The mask in form of n x m array.
        """
        if self.mode_ == 'target':
            return np.apply_along_axis(self._target, 1, np.reshape(X, (X.shape[0], X.shape[1] * X.shape[2])))

        if self.mode_ == 'majority':
            return np.apply_along_axis(self._majority, 1, np.reshape(X, (X.shape[0], X.shape[1] * X.shape[2])))

        print('Invalid mode! Set mode to majority or target. Returning input.')
        return X


class Mask2TwoClass:
    """
    The masks can include non-exclusive-multi-class labels described with bit pattern.
    This transformer simplifies the mask in form of bit-pattern to two class labels.

    Parameters
    ----------
    positive_class_definition: int or string
        int or string (bit pattern, i.e. '100001') defining the positive class
        if argument is an int then the mask is not interpreted as bit pattern
    """

    def __init__(self, positive_class_definition):

        if isinstance(positive_class_definition, str):
            self.definition_ = int(positive_class_definition, 2)
            self.binary_ = True
        else:
            self.definition_ = positive_class_definition
            self.binary_ = False

    def transform(self, X):
        """
        Parameters
        ----------
        X : array-like, shape [n x m]
            The mask in form of n x m array.
        """

        if self.binary_:
            return (X & self.definition_ > 0).astype(int)
        return (X == self.definition_).astype(int)



"""
Tasks used for train set preparation

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc (Sinergise)
Copyright (c) 2017-2019 Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

from enum import Enum
import numpy as np

from eolearn.core import EOTask, FeatureType


class TrainTestSplitType(Enum):
    """ An enum defining TrainTestSplitTask's methods of splitting the data into subsets
    """
    PER_PIXEL = 'per_pixel'
    PER_CLASS = 'per_class'
    PER_VALUE = 'per_value'


class TrainTestSplitTask(EOTask):
    """ Randomly assign each pixel or groups of pixels to multiple subsets (e.g., test/train/validate).

    Input pixels are defined by an input feature (e.g., MASK_TIMELESS with polygon ids, connected component ids, or
    similar), that groups together pixels with similar properties.

    There are three modes of split operation:

    - :attr:`PER_PIXEL<eolearn.ml_tools.train_test_split.TrainTestSplitType.PER_PIXEL>` (default), where pixels are
      assigned to a subset randomly, regardless of their value,

    - :attr:`PER_CLASS<eolearn.ml_tools.train_test_split.TrainTestSplitType.PER_CLASS>`, where pixels of the same value
      are assigned to the same subset,

    - :attr:`PER_VALUE<eolearn.ml_tools.train_test_split.TrainTestSplitType.PER_VALUE>`, where pixels of the same value
      are assigned to a the same subset consistently across eopatches. In other words, if a group of pixels of the same
      value lies on multiple eopatches, they are assigned to the same subset in all eopatches. In this case, the *seed*
      argument of the *execute* method is ignored.

    Classes are defined by a list of cumulative probabilities, passed as the *bins* argument, the same way as the *bins*
    argument in `numpy.digitize <https://docs.scipy.org/doc/numpy/reference/generated/numpy.digitize.html>`_. Valid
    classes are enumerated from 1 onward and if no_data_value is provided, all values equal to it get assigned to class
    0.

    To get a train/test split as 80/20, bins argument should be provided as `bins=[0.8]`.

    To get a train/val/test split as 60/20/20, bins argument should be provided as `bins=[0.6, 0.8]`.

    Splits can also be made into as many subsets as desired, e.g., `bins=[0.1, 0.2, 0.3, 0.7, 0.9]`.

    After the execution of this task an EOPatch will have a new (FeatureType, new_name) feature where each pixel will
    have a value representing the train, test and/or validation set.
    """

    def __init__(self, feature, bins, split_type=TrainTestSplitType.PER_PIXEL, ignore_values=None):
        """
        :param feature: The input feature out of which to generate the train mask.
        :type feature: (FeatureType, feature_name, new_name)
        :param bins: Cumulative probabilities of all value classes or a single float, representing a fraction.
        :type bins: a float or list of floats
        :param split_type: Valye split type, either 'per_pixel', 'per_class' or 'per_value'.
        :type split_type: str
        :param ignore_values: A list of values to ignore and not assign them to any subsets.
        :type ignore_values: a list of integers
        """
        allowed_types = [FeatureType.MASK_TIMELESS]
        self.feature = next(self._parse_features(feature, new_names=True, allowed_feature_types=allowed_types)())

        if np.isscalar(bins):
            bins = [bins]

        if not isinstance(bins, list) or not all(isinstance(bi, float) for bi in bins) \
                or np.any(np.diff(bins) <= 0) or bins[0] <= 0 or bins[-1] >= 1:
            raise ValueError('bins argument should be a list of ascending floats inside an open interval (0, 1)')

        self.ignore_values = set() if ignore_values is None else set(ignore_values)
        self.bins = bins
        self.split_type = TrainTestSplitType(split_type)

    def execute(self, eopatch, *, seed=None):
        """
        :param eopatch: input EOPatch
        :type eopatch: EOPatch
        :param seed: An argument to be passed to numpy.random.seed function.
        :type seed: numpy.int64
        :return: Input EOPatch with the train set mask.
        :rtype: EOPatch
        """
        if self.split_type in [TrainTestSplitType.PER_CLASS, TrainTestSplitType.PER_PIXEL]:
            np.random.seed(seed)

        ftype, fname, new_name = self.feature
        data = eopatch[(ftype, fname)]

        if self.split_type == TrainTestSplitType.PER_PIXEL:
            rands = np.random.rand(*data.shape)
            output_mask = np.digitize(rands, self.bins) + 1
        else:
            classes = set(np.unique(data)) - self.ignore_values
            output_mask = np.zeros_like(data)

            for class_id in classes:
                if self.split_type == TrainTestSplitType.PER_VALUE:
                    np.random.seed(class_id)

                fold = np.digitize(np.random.rand(), self.bins) + 1
                output_mask[data == class_id] = fold

        eopatch[ftype][new_name] = output_mask

        return eopatch



"""
Module for validating results obtained from any ML classifier

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import pickle
import itertools
from abc import ABC, abstractmethod

import numpy as np
import seaborn as sns
import pandas as pd


class SGMLBaseValidator(ABC):
    """
    Abstract class for various validations of SGML image classifiers.

    All the work is performed in the validate method, where
    for each EOPatch the following actions are performed:
        - 1. execute WOWorkflow on EOPatch
        - 2. extract ground truth (reference) from the EOPatch
            - sets self.truth_masks
            - the class values in ground truth has to the same as
              in the provided dictionary
        - 3. count truth labeled pixels
            - sets self.pixel_truth_counts
        - 4. extract classification from the EOPatch
            - sets self.classification_masks
        - 6. count classified pixels
            - sets self.pixel_classification_counts
        - 7. collect results
            - sets pixel_truth_sum and pixel_classification_sum

    Parameters:
    -----------

    workflow: EOWorkflow that is executed on EOPatches (can be simply load EOPacth)

    class_dictionary: dictionary
        Dictionary of class names and class values.

    validation_dirs: list
        List of directories containing EOPatches of this validation sample
    """

    def __init__(self, class_dictionary):
        self.class_dictionary = class_dictionary

        self.n_validation_sets = 0

        self.truth_masks = None
        self.classification_masks = None
        self.pixel_truth_counts = None
        self.pixel_classification_counts = None

        self.pixel_truth_sum = None
        self.pixel_classification_sum = None

        self.truth_classes = None

        self.val_df = None

        super().__init__()

    @abstractmethod
    def _transform_truth(self, patch):
        """
        Transform and extract the truth mask form the EOPatch and store the transformed masks in self.truth_masks.

        Parameters:
        -----------

        patch: EOPatch containing ground truth
        """

    def reset_counters(self):
        """
        Resets all counters, truth and classification masks.
        """
        self.truth_masks = None
        self.classification_masks = None
        self.pixel_truth_counts = None
        self.pixel_classification_counts = None
        self.pixel_truth_sum = None
        self.pixel_classification_sum = None
        self.val_df = None
        self.n_validation_sets = 0

    def _count_truth_pixels(self):
        """
        Count the pixels belonging to each truth class
        """
        pixel_count = np.array([[np.nonzero(mask)[0].shape[0] for mask in masktype]
                                for masktype in self.truth_masks])

        pixel_count = np.moveaxis(pixel_count, 0, -1)

        if self.pixel_truth_counts is None:
            self.pixel_truth_counts = np.copy(pixel_count)
        else:
            self.pixel_truth_counts = np.concatenate((self.pixel_truth_counts, pixel_count))

    def _count_classified_pixels(self):
        """
        Count the pixels belonging to each classified class.
        """
        class_values = self.class_dictionary.values()

        classification_count = np.array([[[np.count_nonzero(prediction[np.nonzero(mask)] == class_val)
                                           for prediction, mask in zip(self.classification_masks, masktype)]
                                          for masktype in self.truth_masks]
                                         for class_val in class_values])

        classification_count = np.moveaxis(classification_count, 0, -1)
        classification_count = np.moveaxis(classification_count, 0, -2)

        if self.pixel_classification_counts is None:
            self.pixel_classification_counts = np.copy(classification_count)
        else:
            self.pixel_classification_counts = np.concatenate((self.pixel_classification_counts, classification_count))

    @abstractmethod
    def _classify(self, patch):
        """
        Extract classification from the EOPatch.

        The classification results should be collected in self.classification_masks
        """

    def add_validation_patch(self, patch):
        """
        Extracts ground truth and classification results from the EOPatch and
        aggregates the results.
        """
        # 2. Convert 8-bit mask
        self._transform_truth(patch)

        # 3. Count truth labeled pixels
        self._count_truth_pixels()

        # 5. Perform classification
        self._classify(patch)

        # 6. Count pixel classified as class i
        self._count_classified_pixels()

        self.n_validation_sets = self.n_validation_sets + 1

    def validate(self):
        """
        Aggregate the results from all EOPatches.
        """
        self.pixel_truth_sum = np.sum(self.pixel_truth_counts, axis=0)
        self.pixel_classification_sum = np.sum(self.pixel_classification_counts, axis=0)

    def save(self, filename):
        """
        Save validator object to pickle.
        """
        with open(filename, 'wb') as output:
            pickle.dump(self, output, protocol=pickle.HIGHEST_PROTOCOL)

    def pandas_df(self):
        """
        Returns pandas DataFrame containing pixel counts for all truth classes,
        classified classes (for each truth class), and file name of the input
        EODataSet.

        The data frame thus contains

        N =  self.n_validation_sets rows

        and

        M = len(self.truth_classes) + len(self.truth_classes) * len (self.class_dictionary) + 1 columns
        """

        if self.val_df is not None:
            return self.val_df

        clf = self.pixel_classification_counts.reshape(self.pixel_classification_counts.shape[0],
                                                       self.pixel_classification_counts.shape[1] *
                                                       self.pixel_classification_counts.shape[2])

        combo = np.hstack((self.pixel_truth_counts, clf))

        columns = list(itertools.product(self.truth_classes, list(self.class_dictionary.keys())))
        columns = [(item[0] + '_as_' + item[1]).replace(" ", "_") for item in columns]
        truth_columns = ['truth_' + item.replace(" ", "_") for item in self.truth_classes]

        self.val_df = pd.DataFrame(combo, columns=truth_columns + columns)

        return self.val_df

    def to_csv(self, filename):
        """
        Writes validation results using pandas to csv file.
        """
        self.pandas_df().to_csv(filename)

    def confusion_matrix(self):
        """
        Returns the normalised confusion matrix
        """
        confusion_matrix = self.pixel_classification_sum.astype(np.float)
        confusion_matrix = np.divide(confusion_matrix.T, self.pixel_truth_sum.T).T

        return confusion_matrix * 100.0

    def plot_confusion_matrix(self, normalised=True):
        """
        Plots the confusion matrix.
        """
        conf_matrix = self.confusion_matrix()

        if normalised:
            sns.heatmap(conf_matrix,
                        annot=True, annot_kws={"size": 12}, fmt='2.1f', cmap='YlGnBu', vmin=0.0,
                        vmax=100.0,
                        xticklabels=list(self.class_dictionary.keys()),
                        yticklabels=self.truth_classes)
        else:
            sns.heatmap(self.pixel_classification_counts,
                        annot=True, annot_kws={"size": 12}, fmt='2.1f', cmap='YlGnBu', vmin=0.0,
                        vmax=np.max(self.pixel_classification_counts),
                        xticklabels=list(self.class_dictionary.keys()),
                        yticklabels=self.truth_classes)

    def summary(self, scoring):
        """
        Prints out the summary of validation for giving scoring function.
        """

        if scoring == 'class_confusion':
            print('*' * 50)
            print('  Confusion Matrix ')
            print('x-axis: ' + ' | '.join(list(self.class_dictionary.keys())))
            print('y-axis: ' + ' | '.join(self.truth_classes))
            print(self.confusion_matrix())



"""
Module contains tools and EOTasks for morphological and postprocessing operations

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import itertools as it
from enum import Enum
from abc import abstractmethod

import skimage.morphology
import skimage.filters.rank

from eolearn.core import EOTask


class MorphologicalOperations(Enum):
    """ Enum class of morphological operations
    """
    OPENING = 'opening'
    CLOSING = 'closing'
    DILATION = 'dilation'
    EROSION = 'erosion'
    MEDIAN = 'median'

    @classmethod
    def get_operation(cls, morph_type):
        """ Maps morphological operation type to function

        :param morph_type: Morphological operation type
        :type morph_type: MorphologicalOperations
        :return: function
        """
        return {
            cls.OPENING: skimage.morphology.opening,
            cls.CLOSING: skimage.morphology.closing,
            cls.DILATION: skimage.morphology.dilation,
            cls.EROSION: skimage.morphology.erosion,
            cls.MEDIAN: skimage.filters.rank.median
        }[morph_type]


class MorphologicalStructFactory:
    """
    Factory methods for generating morphological structuring elements
    """
    @staticmethod
    def get_disk(radius):
        """
        :param radius: Radius of disk
        :type radius: int
        :return: The structuring element where elements of the neighborhood are 1 and 0 otherwise.
        :rtype: numpy.ndarray
        """
        return skimage.morphology.disk(radius)

    @staticmethod
    def get_diamond(radius):
        """
        :param radius: Radius of diamond
        :type radius: int
        :return: The structuring element where elements of the neighborhood are 1 and 0 otherwise.
        :rtype: numpy.ndarray
        """
        return skimage.morphology.diamond(radius)

    @staticmethod
    def get_rectangle(width, height):
        """
        :param width: Width of rectangle
        :type width: int
        :param height: Height of rectangle
        :type height: int
        :return: A structuring element consisting only of ones, i.e. every pixel belongs to the neighborhood.
        :rtype: numpy.ndarray
        """
        return skimage.morphology.rectangle(width, height)

    @staticmethod
    def get_square(width):
        """
        :param width: Size of square
        :type width: int
        :return: A structuring element consisting only of ones, i.e. every pixel belongs to the neighborhood.
        :rtype: numpy.ndarray
        """
        return skimage.morphology.square(width)


class PostprocessingTask(EOTask):
    """ Base class for all post-processing tasks

    :param feature: A feature to be processed
    :type feature: (FeatureType, str)
    """
    def __init__(self, feature):
        self.feature = self._parse_features(feature)

    @abstractmethod
    def process(self, raster):
        """ Abstract method for processing the raster
        """
        raise NotImplementedError

    def execute(self, eopatch):
        """ Execute method takes EOPatch and changes the specified feature
        """
        feature_type, feature_name = next(self.feature(eopatch))

        eopatch[feature_type][feature_name] = self.process(eopatch[feature_type][feature_name])

        return eopatch


class MorphologicalFilterTask(PostprocessingTask):
    """ EOTask that performs morphological operations on masks.

    :param feature: A feature to be processed
    :type feature: (FeatureType, str)
    :param morph_operation: Morphological operation
    :type morph_operation: MorphologicalOperations or function that operates on image
    :param struct_elem: The structuring element to be used with the morphological operation; usually generated with a
                        factory method from MorphologicalStructElements
    :type struct_elem: numpy.ndarray
    """
    def __init__(self, feature, morph_operation, struct_elem=None):
        super().__init__(feature)

        if isinstance(morph_operation, MorphologicalOperations):
            self.morph_operation = MorphologicalOperations.get_operation(morph_operation)
        else:
            self.morph_operation = morph_operation
        self.struct_elem = struct_elem

    def process(self, raster):
        """ Applies the morphological operation to the mask object
        """
        dim = len(raster.shape)
        if dim == 3:
            for dim in range(raster.shape[2]):
                raster[:, :, dim] = self.morph_operation(raster[:, :, dim], self.struct_elem)
        elif dim == 4:
            for time, dim in it.product(range(raster.shape[0]), range(raster.shape[3])):
                raster[time, :, :, dim] = self.morph_operation(raster[time, :, :, dim], self.struct_elem)
        else:
            raise ValueError('Invalid number of dimensions: {}'.format(dim))

        return raster



"""
The utilities module is a collection of methods used across the eolearn package, such as checking whether two objects
are deeply equal, padding of an image, etc.

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging
import itertools

import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    import matplotlib
    matplotlib.use('agg')
    import matplotlib.pyplot as plt


LOGGER = logging.getLogger(__name__)


# This code was copied from https://gist.github.com/seberg/3866040
def rolling_window(array, window=(0,), asteps=None, wsteps=None, axes=None, toend=True):
    """Create a view of `array` which for every point gives the n-dimensional
    neighbourhood of size window. New dimensions are added at the end of
    `array` or after the corresponding original dimension.

    Parameters
    ----------
    array : array_like
        Array to which the rolling window is applied.
    window : int or tuple
        Either a single integer to create a window of only the last axis or a
        tuple to create it for the last len(window) axes. 0 can be used as a
        to ignore a dimension in the window.
    asteps : tuple
        Aligned at the last axis, new steps for the original array, ie. for
        creation of non-overlapping windows. (Equivalent to slicing result)
    wsteps : int or tuple (same size as window)
        steps for the added window dimensions. These can be 0 to repeat values
        along the axis.
    axes: int or tuple
        If given, must have the same size as window. In this case window is
        interpreted as the size in the dimension given by axes. IE. a window
        of (2, 1) is equivalent to window=2 and axis=-2.
    toend : bool
        If False, the new dimensions are right after the corresponding original
        dimension, instead of at the end of the array. Adding the new axes at the
        end makes it easier to get the neighborhood, however toend=False will give
        a more intuitive result if you view the whole array.

    Returns
    -------
    A view on `array` which is smaller to fit the windows and has windows added
    dimensions (0s not counting), ie. every point of `array` is an array of size
    window.

    Examples
    --------
    >>> a = np.arange(9).reshape(3,3)
    >>> rolling_window(a, (2,2))
    array([[[[0, 1],
             [3, 4]],

            [[1, 2],
             [4, 5]]],


           [[[3, 4],
             [6, 7]],

            [[4, 5],
             [7, 8]]]])

    Or to create non-overlapping windows, but only along the first dimension:
    >>> rolling_window(a, (2,0), asteps=(2,1))
    array([[[0, 3],
            [1, 4],
            [2, 5]]])

    Note that the 0 is discared, so that the output dimension is 3:
    >>> rolling_window(a, (2,0), asteps=(2,1)).shape
    (1, 3, 2)

    This is useful for example to calculate the maximum in all (overlapping)
    2x2 submatrixes:
    >>> rolling_window(a, (2,2)).max((2,3))
    array([[4, 5],
           [7, 8]])

    Or delay embedding (3D embedding with delay 2):
    >>> x = np.arange(10)
    >>> rolling_window(x, 3, wsteps=2)
    array([[0, 2, 4],
           [1, 3, 5],
           [2, 4, 6],
           [3, 5, 7],
           [4, 6, 8],
           [5, 7, 9]])
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements

    array = np.asarray(array)
    orig_shape = np.asarray(array.shape)
    window = np.atleast_1d(window).astype(int)  # maybe crude to cast to int...

    if axes is not None:
        axes = np.atleast_1d(axes)
        new_window = np.zeros(array.ndim, dtype=int)
        for axis, size in zip(axes, window):
            new_window[axis] = size
        window = new_window

    # Check if window is legal:
    if window.ndim > 1:
        raise ValueError("`window` must be one-dimensional.")
    if np.any(window < 0):
        raise ValueError("All elements of `window` must be larger then 1.")
    if len(array.shape) < len(window):
        raise ValueError("`window` length must be less or equal `array` dimension.")

    _asteps = np.ones_like(orig_shape)
    if asteps is not None:
        asteps = np.atleast_1d(asteps)
        if asteps.ndim != 1:
            raise ValueError("`asteps` must be either a scalar or one dimensional.")
        if len(asteps) > array.ndim:
            raise ValueError("`asteps` cannot be longer then the `array` dimension.")
        # does not enforce alignment, so that steps can be same as window too.
        _asteps[-len(asteps):] = asteps

        if np.any(asteps < 1):
            raise ValueError("All elements of `asteps` must be larger then 1.")
    asteps = _asteps

    _wsteps = np.ones_like(window)
    if wsteps is not None:
        wsteps = np.atleast_1d(wsteps)
        if wsteps.shape != window.shape:
            raise ValueError("`wsteps` must have the same shape as `window`.")
        if np.any(wsteps < 0):
            raise ValueError("All elements of `wsteps` must be larger then 0.")

        _wsteps[:] = wsteps
        _wsteps[window == 0] = 1  # make sure that steps are 1 for non-existing dims.
    wsteps = _wsteps

    # Check that the window would not be larger then the original:
    if np.any(orig_shape[-len(window):] < window * wsteps):
        raise ValueError("`window` * `wsteps` larger then `array` in at least one dimension.")

    new_shape = orig_shape  # just renaming...

    # For calculating the new shape 0s must act like 1s:
    _window = window.copy()
    _window[_window == 0] = 1

    new_shape[-len(window):] += wsteps - _window * wsteps
    new_shape = (new_shape + asteps - 1) // asteps
    # make sure the new_shape is at least 1 in any "old" dimension (ie. steps
    # is (too) large, but we do not care.
    new_shape[new_shape < 1] = 1
    shape = new_shape

    strides = np.asarray(array.strides)
    strides *= asteps
    new_strides = array.strides[-len(window):] * wsteps

    # The full new shape and strides:
    if toend:
        new_shape = np.concatenate((shape, window))
        new_strides = np.concatenate((strides, new_strides))
    else:
        _ = np.zeros_like(shape)
        _[-len(window):] = window
        _window = _.copy()
        _[-len(window):] = new_strides
        _new_strides = _

        new_shape = np.zeros(len(shape) * 2, dtype=int)
        new_strides = np.zeros(len(shape) * 2, dtype=int)

        new_shape[::2] = shape
        new_strides[::2] = strides
        new_shape[1::2] = _window
        new_strides[1::2] = _new_strides

    new_strides = new_strides[new_shape != 0]
    new_shape = new_shape[new_shape != 0]

    return np.lib.stride_tricks.as_strided(array, shape=new_shape, strides=new_strides)


def plot_confusion_matrix(matrix, classes, normalize=False, title='Confusion matrix',
                          cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    (This function is copied from the scikit docs.)
    """
    plt.figure()
    plt.imshow(matrix, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = '.2f' if normalize else 'd'
    if normalize:
        matrix = matrix.astype('float') / matrix.sum(axis=1)[:, np.newaxis]
    print(matrix)
    thresh = matrix.max() / 2.
    for i, j in itertools.product(range(matrix.shape[0]), range(matrix.shape[1])):
        plt.text(j, i, format(matrix[i, j], fmt), horizontalalignment="center",
                 color="white" if matrix[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')



"""
Module for classification helper classes and classification task.

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
# pylint: disable=invalid-name

import itertools
from abc import ABC, abstractmethod

import numpy as np

from eolearn.core import EOTask

from .utilities import rolling_window


class ImageBaseClassifier(ABC):
    """
    Abstract class for image classifiers.

    Image Classifier extends the receptive field of trained classifier with smaller
    receptive field over entire image. The classifier's receptive field is
    usually small, i.e.:
        - pixel based classifier has receptive field (1,1)
        - patch based classifier has receptive field (num_pixels_y, num_pixels_x)

    Image Classifier divides the image into non-overlapping pieces of same size
    as trained classifier's receptive field and runs classifier over them thus
    producing a classification mask of the same size as image.

    The classifier can be of any type as long as it has the following two
    methods implemented:
        - predict(X)
        - predict_proba(X)

    This is true for all classifiers that follow scikit-learn's API.
    The APIs of scikit-learn's objects is described
    at: http://scikit-learn.org/stable/developers/contributing.html#apis-of-scikit-learn-objects.

    Parameters:
    -----------

    classifier:
        The actual trained classifier that will be executed over entire image

    receptive_field: tuple, (n_rows, n_columns)
        Sensitive area of the classifier ((1,1) for pixel based or (n,m) for patch base)

    """

    def __init__(self, classifier, receptive_field):
        self.receptive_field = receptive_field

        self._check_classifier(classifier)
        self.classifier = classifier

        self._samples = None
        self._image_size = None

    @staticmethod
    def _check_classifier(classifier):
        """
        Check if the classifier implements predict and predict_proba methods.
        """
        predict = getattr(classifier, "predict", None)
        if not callable(predict):
            raise ValueError('Classifier does not have predict method!')

        predict_proba = getattr(classifier, "predict_proba", None)
        if not callable(predict_proba):
            raise ValueError('Classifier does not have predict_proba method!')

    def _check_image(self, X):
        """
        Checks the image size and its compatibility with classifier's receptive field.

        At this moment it is required that image size = K * receptive_field. This will
        be relaxed in future with the introduction of padding.
        """

        if (len(X.shape) < 3) or (len(X.shape) > 4):
            raise ValueError('Input has to have shape [n_samples, n_pixels_y, n_pixels_x] '
                             'or [n_samples, n_pixels_y, n_pixels_x, n_bands].')

        self._samples = X.shape[0]
        self._image_size = X.shape[1:3]

        if (self._image_size[0] % self.receptive_field[0]) or (self._image_size[0] % self.receptive_field[0]):
            raise ValueError('Image (%d,%d) and receptive fields (%d,%d) mismatch.\n'
                             'Resize your image to be divisible with receptive field.'
                             % (self._image_size[0], self._image_size[0], self.receptive_field[0],
                                self.receptive_field[1]))

    @staticmethod
    def _transform_input(X):
        """
        Transform the input in the form expected by the classifier. For example reshape matrix to vector.
        """
        return X

    @abstractmethod
    def image_predict(self, X):
        """
        Predicts class label for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_y, n_pixels_x, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_y, n_pixels_x]
            Target labels or masks.
        """
        raise NotImplementedError

    @abstractmethod
    def image_predict_proba(self, X):
        """
        Predicts class probabilities for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_x, n_pixels_y, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_x, n_pixels_y, n_classes]
            Target probabilities
        """
        raise NotImplementedError


class ImagePixelClassifier(ImageBaseClassifier):
    """
    Pixel classifier divides the image into individual pixels,
    runs classifier and collects the result in the shape of the input image.

    Parameters:
    -----------

    classifier:
        The actual trained classifier that will be executed over entire image
    """

    def __init__(self, classifier):
        ImageBaseClassifier.__init__(self, classifier, (1, 1))

    def image_predict(self, X):
        """
        Predicts class label for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_y, n_pixels_x, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_y, n_pixels_x]
            Target labels or masks.
        """
        self._check_image(X)

        new_shape = (X.shape[0] * X.shape[1] * X.shape[2],)

        if len(X.shape) == 4:
            new_shape += (X.shape[3],)

        pixels = X.reshape(new_shape)

        predictions = self.classifier.predict(self._transform_input(pixels))

        return predictions.reshape(X.shape[0], X.shape[1], X.shape[2])

    def image_predict_proba(self, X):
        """
        Predicts class probabilities for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_x, n_pixels_y, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_x, n_pixels_y, n_classes]
            Target probabilities
        """
        self._check_image(X)

        new_shape = (X.shape[0] * X.shape[1] * X.shape[2],)

        if len(X.shape) == 4:
            new_shape += (X.shape[3],)

        pixels = X.reshape(new_shape)

        probabilities = self.classifier.predict_proba(self._transform_input(pixels))

        return probabilities.reshape(X.shape[0], X.shape[1], X.shape[2],
                                     probabilities.shape[1])


class ImagePatchClassifier(ImageBaseClassifier):
    """
    Patch classifier divides the image into non-overlapping patches of same size
    as trained classifier's receptieve field and runs classifier over them thus
    producing a classification mask of the same size as image.

    Parameters:
    -----------

    classifier:
        The actual trained classifier that will be executed over entire image

    receptive_field: tuple, (n_rows, n_columns)
        Sensitive area of the classifier ((1,1) for pixel based or (n,m) for patch based)

    """

    def _to_patches(self, X):
        """
        Reshapes input to patches of the size of classifier's receptive field.

        For example:

        input X shape: [n_samples, n_pixels_y, n_pixels_x, n_bands]

        output: [n_samples * n_pixels_y/receptive_field_y * n_pixels_x/receptive_field_x,
                 receptive_field_y, receptive_field_x, n_bands]
        """

        window = self.receptive_field
        asteps = self.receptive_field

        if len(X.shape) == 4:
            window += (0,)
            asteps += (1,)

        image_view = rolling_window(X, window, asteps)

        new_shape = image_view.shape

        # this makes a copy of the array? can we do without reshaping?
        image_view = image_view.reshape((new_shape[0] * new_shape[1] * new_shape[2],) + new_shape[3:])

        if len(X.shape) == 4:
            image_view = np.moveaxis(image_view, 1, -1)

        return image_view, new_shape

    def image_predict(self, X):
        """
        Predicts class label for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_y, n_pixels_x, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_y, n_pixels_x]
            Target labels or masks.
        """
        self._check_image(X)

        patches, patches_shape = self._to_patches(X)

        predictions = self.classifier.predict(self._transform_input(patches))

        image_predictions = predictions.reshape(patches_shape[0:3])

        image_results = np.zeros((self._samples,) + self._image_size)

        nx, ny = self.receptive_field
        row_steps = self._image_size[0] // nx
        col_steps = self._image_size[1] // ny

        # how can this be optimised?
        for i, j, k in itertools.product(range(row_steps), range(col_steps), range(self._samples)):
            image_results[k, nx * i:nx * (i + 1), ny * j:ny * (j + 1)] = image_predictions[k, i, j]

        return image_results

    def image_predict_proba(self, X):
        """
        Predicts class probabilities for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_x, n_pixels_y, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_x, n_pixels_y, n_classes]
            Target probabilities
        """
        self._check_image(X)

        patches, patches_shape = self._to_patches(X)

        probabilities = self.classifier.predict_proba(self._transform_input(patches))

        image_probabilities = probabilities.reshape(patches_shape[0:3] + (probabilities.shape[1],))

        image_results = np.zeros((self._samples,) + self._image_size + (probabilities.shape[1],))

        nx, ny = self.receptive_field
        row_steps = self._image_size[0] // nx
        col_steps = self._image_size[1] // ny

        # how can this be optimised?
        for i, j, k in itertools.product(range(row_steps), range(col_steps), range(self._samples)):
            image_results[k, nx * i:nx * (i + 1), ny * j:ny * (j + 1), :] = image_probabilities[k, i, j, :]

        return image_results


class ImagePixel2PatchClassifier(ImageBaseClassifier):
    """
    Pixel to patch classifier first performs classification on pixel level
    and then combines the results in user defined patches. In case of combining
    probabilities the weighted sum is taken over all pixels in a patch. In case
    of predictions the user defines what fraction of pixels within the patch
    has to belong to signal class ot be considered as signal.

    Parameters:
    -----------

    classifier:
        The actual trained classifier that will be executed over entire image

    patch_size: tuple, (n_rows, n_columns)
        Patch size

    target: int
        Target class value. Set the patch class to this target class if its fractional representation within
        this patch is above the target_threshols

    target_threshold: float
        See above

    """

    def __init__(self, classifier, patch_size, mode='mean_prob', target=None, target_threshold=None):
        self.pixel_classifier = ImagePixelClassifier(classifier)
        self.patch_size = patch_size
        self.target = target
        self.target_threshold = target_threshold

        self.mode = mode

        ImageBaseClassifier.__init__(self, classifier, (1, 1))

    def _to_patches(self, X):
        """
        Reshapes input to patches of the size of classifier's receptive field.

        For example:

        input X shape: [n_samples, n_pixels_y, n_pixels_x, n_bands]

        output: [n_samples * n_pixels_y/receptive_field_y * n_pixels_x/receptive_field_x,
                 receptive_field_y, receptive_field_x, n_bands]
        """

        window = self.patch_size
        asteps = self.patch_size

        if len(X.shape) == 4:
            window += (0,)
            asteps += (1,)

        image_view = rolling_window(X, window, asteps)

        new_shape = image_view.shape

        return image_view, new_shape

    def _target(self, array):
        unique, counts = np.unique(array, return_counts=True)
        valuecount = dict(zip(unique, counts))

        return 1 if self.target in valuecount.keys() and \
                    valuecount[self.target] / np.ma.size(array) >= self.target_threshold else 0

    def image_predict(self, X):
        """
        Predicts class label for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_y, n_pixels_x, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_y, n_pixels_x]
            Target labels or masks.
        """
        self._check_image(X)

        if self.mode == 'majority_class':
            predictions = self.pixel_classifier.image_predict(X)

        elif self.mode == 'mean_prob':
            probabilities = self.image_predict_proba(X)
            predictions = (probabilities[..., self.target] > self.target_threshold).astype(np.int)

        patches, _ = self._to_patches(predictions)

        row_steps = self._image_size[0] // self.patch_size[0]
        col_steps = self._image_size[1] // self.patch_size[1]

        # how can this be optimised?
        for i, j, k in itertools.product(range(row_steps), range(col_steps), range(self._samples)):
            patches[k, i, j] = self._target(patches[k, i, j])

        return predictions

    def image_predict_proba(self, X):
        """
        Predicts class probabilities for the entire image.

        Parameters:
        -----------

        X: array, shape = [n_samples, n_pixels_x, n_pixels_y, n_bands]
            Array of training images

        y: array, shape = [n_samples] or [n_samples, n_pixels_x, n_pixels_y, n_classes]
            Target probabilities
        """
        self._check_image(X)

        probabilities = self.pixel_classifier.image_predict_proba(X)

        patches, _ = self._to_patches(probabilities)

        row_steps = self._image_size[0] // self.patch_size[0]
        col_steps = self._image_size[1] // self.patch_size[1]

        ps = self.patch_size[0] * self.patch_size[1]

        # how can this be optimised?
        for i, j, k in itertools.product(range(row_steps), range(col_steps), range(self._samples)):
            patches[k, i, j, 0] = np.sum(patches[k, i, j, 0]) / ps
            patches[k, i, j, 1] = np.sum(patches[k, i, j, 1]) / ps

        return probabilities


class ImageClassificationMaskTask(EOTask):
    """
    This task applies pixel-based uni-temporal classifier to each image in the patch
    and appends to each image the classification mask.
    """
    def __init__(self, input_feature, output_feature, classifier):
        """ Run a classification task on a EOPatch feature

            Classifier is an instance of the ImageBaseClassifier that maps [w, h, d] numpy arrays (d-channel images)
            into [w, h, 1] numpy arrays (classification masks).

            :param input_feature: Feature which will be classified
            :type input_feature: (FeatureType, str)
            :param output_feature: Feature where classification results will be saved
            :type output_feature: (FeatureType, str)
            :param classifier: A classifier that works over [n, w, h, d]-dimensional numpy arrays.
            :type classifier: ImageBaseClassifier
        """
        self.input_feature = self._parse_features(input_feature)
        self.output_feature = self._parse_features(output_feature)
        self.classifier = classifier

    def execute(self, eopatch):
        """ Transforms [n, w, h, d] eopatch into a [n, w, h, 1] eopatch, adding it the classification mask.

            :param eopatch: An input EOPatch
            :type eopatch: EOPatch
            :return: Outputs EOPatch with n classification masks appended to out_feature_type with out_feature_name key
            :rtype: EOPatch
        """
        in_type, in_name = next(self.input_feature(eopatch))
        out_type, out_name = next(self.input_feature())

        eopatch[out_type][out_name] = self.classifier.image_predict(eopatch[in_type][in_name])

        return eopatch
